"""Utility functions.

This file contains miscellaneous useful functions.
"""
from os import path
from typing import Union
from collections.abc import MutableMapping, Mapping
import numpy as np


def safe_path(arg, default, valid_extensions=('.yaml', '.py')):

    if path.isfile(arg):
        if arg.endswith(valid_extensions):
            return path.abspath(arg)
        else:
            raise TypeError("Unsupported file extension. \n"
                            "Valid file extensions: {}".format(arg, valid_extensions))
    elif path.isdir(arg):
        return path.join(arg, default)
    else:
        raise FileNotFoundError("Directory or file ({}) not found.".format(arg))


def quasirand(ndim=1, npoint=1):
    from .halton import halton
    return halton(npoint, ndim)


def check_ndim(arr):
    return arr if arr.ndim > 1 else arr.reshape(-1, 1)


class SafeDict(dict):
    def __init__(self, obj, pre='{', post='}'):
        self.pre = pre
        self.post = post
        super().__init__(obj)

    @classmethod
    def from_params(cls, params, **kwargs):
        return cls(params2map(params), **kwargs)

    def __missing__(self, key):
        return self.pre + key + self.post


def params2map(params: Union[None, MutableMapping, np.ndarray, np.void]):
    if params is None:
        return {}
    if isinstance(params, MutableMapping):
        return params
    try:
        return {key: params[key] for key in params.dtype.names}
    except AttributeError:
        pass
    raise TypeError('params are not a Mapping')


def load_includes(paths):
    """ load python modules from the specified paths """
    import os
    import sys
    from importlib.util import spec_from_file_location, module_from_spec
    import logging
    for path in paths:
        name = f"profit_include_{os.path.basename(path).split('.')[0]}"
        if name in sys.modules:  # do not reload modules
            continue
        try:
            spec = spec_from_file_location(name, path)
        except FileNotFoundError:
            logging.getLogger(__name__).error(f'could not find {path} to include')
            continue
        module = module_from_spec(spec)
        sys.modules[name] = module
        spec.loader.exec_module(module)

def flatten_struct(struct_array: np.ndarray):
    # per default vector entries are spread across several columns
    if not struct_array.size:
        return np.array([[]])
    return np.vstack([np.hstack([row[key].flatten() for key in struct_array.dtype.names]) for row in struct_array])



""" Default Runner & Worker components

Local Runner
Memmap Interface (numpy)
Template Preprocessor
JSON Postprocessor
NumpytxtPostprocessor
HDF5Postprocessor
"""

from .runner import Runner, RunnerInterface
from .worker import Interface, Preprocessor, Postprocessor, Worker

import subprocess
from multiprocessing import Process
from time import sleep
import logging

import numpy as np
import os
from shutil import rmtree

# === Local Runner === #


@Runner.register('local')
class LocalRunner(Runner):
    """ Runner for executing simulations locally

    - forks the worker, thereby having less overhead (especially with a custom python Worker)
    - per default uses all available CPUs
    """
    def spawn_run(self, params=None, wait=False):
        super().spawn_run(params, wait)
        if self.run_config['custom'] or not self.config['fork']:
            env = self.env.copy()
            env['PROFIT_RUN_ID'] = str(self.next_run_id)
            if self.run_config['custom']:
                cmd = self.run_config['command']
            else:
                cmd = 'profit-worker'
            self.runs[self.next_run_id] = subprocess.Popen(cmd, shell=True, env=env, cwd=self.base_config['run_dir'])
            if wait:
                self.runs[self.next_run_id].wait()
                del self.runs[self.next_run_id]
        else:
            def work():
                worker = Worker.from_config(self.run_config, self.next_run_id)
                worker.main()
            os.chdir(self.base_config['run_dir'])
            process = Process(target=work)
            self.runs[self.next_run_id] = process
            process.start()
            if wait:
                process.join()
                del self.runs[self.next_run_id]
            os.chdir(self.base_config['base_dir'])
        self.next_run_id += 1

    def spawn_array(self, params_array, blocking=True):
        """ spawn an array of runs, maximum 'parallel' at the same time, blocking until all are done """
        if not blocking:
            raise NotImplementedError
        for params in params_array:
            self.spawn_run(params)
            while len(self.runs) >= self.config['parallel']:
                sleep(self.config['sleep'])
                self.check_runs(poll=True)
        while len(self.runs):
            sleep(self.config['sleep'])
            self.check_runs(poll=True)

    def check_runs(self, poll=False):
        """ check the status of runs via the interface """
        self.interface.poll()
        if self.run_config['custom'] or not self.config['fork']:
            for run_id, process in list(self.runs.items()):  # preserve state before deletions
                if self.interface.internal['DONE'][run_id]:
                    process.wait()  # just to make sure
                    del self.runs[run_id]
                elif poll and process.poll() is not None:
                    # job has crashed or completed -> check backup
                    if not self.check_backup(run_id):
                        self.failed[run_id] = self.runs[run_id]
                    del self.runs[run_id]
        else:
            for run_id, process in list(self.runs.items()):  # preserve state before deletions
                if self.interface.internal['DONE'][run_id]:
                    process.join()  # just to make sure
                    del self.runs[run_id]
                elif poll and process.exitcode is not None:
                    process.terminate()
                    # job has crashed or completed -> check backup
                    if not self.check_backup(run_id):
                        self.failed[run_id] = self.runs[run_id]
                    del self.runs[run_id]

    def cancel_all(self):
        if self.run_config['custom'] or not self.config['fork']:
            for process in self.runs.values():
                process.terminate()
        else:
            for process in self.runs.values():
                process.terminate()
        self.failed = self.runs
        self.runs = {}


# === Numpy Memmap Interface === #


@RunnerInterface.register('memmap')
class MemmapRunnerInterface(RunnerInterface):
    """ Runner-Worker Interface using a memory mapped numpy array

    - expected to be very fast with the *local* Runner as each Worker can access the array directly (unverified)
    - expected to be inefficient if used on a cluster with a shared filesystem (unverified)
    - reliable
    - known issue: resizing the array (to add more runs) is dangerous, needs a workaround
      (e.g. several arrays in the same file)
    """
    def __init__(self, config, size, input_config, output_config, *, logger_parent: logging.Logger = None):
        super().__init__(config, size, input_config, output_config, logger_parent=logger_parent)

        init_data = np.zeros(size, dtype=self.input_vars + self.internal_vars + self.output_vars)
        np.save(self.config['path'], init_data)

        try:
            self._memmap = np.load(self.config['path'], mmap_mode='r+')
        except FileNotFoundError:
            self.runner.logger.error(
                f'{self.__class__.__name__} could not load {self.config["path"]} (cwd: {os.getcwd()})')
            raise

        # should return views on memmap
        self.input = self._memmap[[v[0] for v in self.input_vars]]
        self.output = self._memmap[[v[0] for v in self.output_vars]]
        self.internal = self._memmap[[v[0] for v in self.internal_vars]]
    
    def resize(self, size):
        """ Resizing Memmap Runner Interfac
        
        Attention: this is dangerous and may lead to unexpected errors!
        The problem is that the memory mapped file is overwritten.
        Any Workers which have this file mapped will run into severe problems.
        Possible future workarounds: multiple files or multiple headers in one file.
        """
        if size <= self.size:
            self.logger.warning('shrinking RunnerInterface is not supported')
            return
        
        self.logger.warning('resizing MemmapRunnerInterface is dangerous')
        self.clean()
        init_data = np.zeros(size, dtype=self.input_vars + self.internal_vars + self.output_vars)
        np.save(self.config['path'], init_data)

        try:
            self._memmap = np.load(self.config['path'], mmap_mode='r+')
        except FileNotFoundError:
            self.runner.logger.error(
                f'{self.__class__.__name__} could not load {self.config["path"]} (cwd: {os.getcwd()})')
            raise

        self.input = self._memmap[[v[0] for v in self.input_vars]]
        self.output = self._memmap[[v[0] for v in self.output_vars]]
        self.internal = self._memmap[[v[0] for v in self.internal_vars]]

    def clean(self):
        if os.path.exists(self.config['path']):
            os.remove(self.config['path'])#


@Interface.register('memmap')
class MemmapInterface(Interface):
    """ Runner-Worker Interface using a memory mapped numpy array

    counterpart to :py:class:`MemmapRunnerInterface`
    """
    def __init__(self, config, run_id: int, *, logger_parent: logging.Logger = None):
        super().__init__(config, run_id, logger_parent=logger_parent)
        # ToDo: multiple arrays after another to allow extending the file dynamically
        try:
            self._memmap = np.load(self.config['path'], mmap_mode='r+')
        except FileNotFoundError:
            self.worker.logger.error(
                f'{self.__class__.__name__} could not load {self.config["path"]} (cwd: {os.getcwd()})')
            raise

        # should return views on memmap
        inputs, outputs = [], []
        k = 0
        for k, key in enumerate(self._memmap.dtype.names):
            if key == 'DONE':
                break
            inputs.append(key)
        for key in self._memmap.dtype.names[k:]:
            if key not in ['DONE', 'TIME']:
                outputs.append(key)
        self.input = self._memmap[inputs][run_id]
        self.output = self._memmap[outputs][run_id]
        self._data = self._memmap[run_id]

    def done(self):
        self._memmap['TIME'] = self.time
        self._memmap['DONE'] = True
        self._memmap.flush()

    def clean(self):
        if os.path.exists(self.config['path']):
            os.remove(self.config['path'])


# === Template Preprocessor === #


@Preprocessor.register('template')
class TemplatePreprocessor(Preprocessor):
    """ Preprocessor which substitutes the variables with a given template

    - copies the given template directory to the target run directory
    - searches all files for variables templates of the form {name} and replaces them with their values
    - for file formats which use curly braces (e.g. json) the template identifier is {{name}}
    - substitution can be restricted to certain files by specifying `param_files`
    - relative symbolic links are converted to absolute symbolic links on copying
    - linked files are ignored with `param_files: all`, but if specified explicitly the link target is copied to the run
      directory and then substituted
    """
    def pre(self, data, run_dir):
        # No call to super()! replaces the default preprocessing
        if os.path.exists(run_dir):
            rmtree(run_dir)
        self.fill_run_dir_single(data, self.config['path'], run_dir, ignore_path_exists=True,
                            param_files=self.config['param_files'])
        os.chdir(run_dir)

    def fill_run_dir_single(self, params, template_dir, run_dir_single, param_files=None, overwrite=False,
                            ignore_path_exists=False):
        if os.path.exists(run_dir_single) and not ignore_path_exists:  # ToDo: make ignore_path_exists default
            if overwrite:
                rmtree(run_dir_single)
            else:
                raise RuntimeError('Run directory not empty: {}'.format(run_dir_single))
        self.copy_template(template_dir, run_dir_single)

        self.fill_template(run_dir_single, params, param_files=param_files)

    def copy_template(self, template_dir, out_dir, dont_copy=None):
        from shutil import copytree, ignore_patterns

        if dont_copy:
            copytree(template_dir, out_dir, symlinks=True, ignore=ignore_patterns(*dont_copy))
        else:
            copytree(template_dir, out_dir, symlinks=True)
        self.convert_relative_symlinks(template_dir, out_dir)

    @staticmethod
    def convert_relative_symlinks(template_dir, out_dir):
        """When copying the template directory to the single run directories,
         relative paths in symbolic links are converted to absolute paths."""
        for root, dirs, files in os.walk(out_dir):
            for filename in files:
                filepath = os.path.join(root, filename)
                if os.path.islink(filepath):
                    linkto = os.readlink(filepath)
                    if linkto.startswith('.'):
                        os.remove(filepath)
                        start_dir = os.path.relpath(root, out_dir)
                        os.symlink(os.path.join(template_dir, start_dir, filename), filepath)

    def fill_template(self, out_dir, params, param_files=None):
        """
        Arguments:
            param_files(list): a list of filenames which are to be substituted or None for all
        """
        if param_files is None:
            param_files = []
        for root, dirs, files in os.walk(out_dir):  # by default, walk ignores subdirectories which are links
            for filename in files:
                filepath = os.path.join(root, filename)
                if (not param_files and not os.path.islink(filepath)) or filename in param_files:
                    self.fill_template_file(filepath, filepath, params)

    def fill_template_file(self, template_filepath, output_filepath, params, copy_link=True):
        """Fill template in `template_filepath` by `params` and output into
        `output_filepath`. If `copy_link` is set (default), do not write into
        symbolic links but copy them instead.
        """
        with open(template_filepath, 'r') as f:
            content = self.replace_template(f.read(), params)
        if copy_link and os.path.islink(output_filepath):
            os.remove(output_filepath)  # otherwise the link target would be substituted
        with open(output_filepath, 'w') as f:
            f.write(content)

    @staticmethod
    def replace_template(content, params):
        """Returns filled template by putting values of `params` in `content`.

        # Escape '{*}' for e.g. json templates by replacing it with '{{*}}'.
        # Variables then have to be declared as '{{*}}' which is replaced by a single '{*}'.
        """
        from profit.util import SafeDict
        pre, post = '{', '}'
        if '{{' in content:
            content = content.replace('{{', '§').replace('}}', '§§') \
                .replace('{', '{{').replace('}', '}}').replace('§§', '}').replace('§', '{')
            pre, post = '{{', '}}'
        return content.format_map(SafeDict.from_params(params, pre=pre, post=post))


# === JSON Postprocessor === #


@Postprocessor.wrap('json')
def JSONPostprocessor(self, data):
    """ Postprocessor to read output from a JSON file

    - variables are assumed to be stored with the correct key and able to be converted immediately
    - not extensively tested
    """
    import json
    with open(self.config['path']) as f:
        output = json.load(f)
    for key, value in output.items():
        data[key] = value


# === Numpy Text Postprocessor === #


@Postprocessor.wrap('numpytxt')
def NumpytxtPostprocessor(self, data):
    """ Postprocessor to read output from a tabular text file (e.g. csv, tsv) with numpy ``genfromtxt``

    - the data is assumed to be row oriented
    - vector variables are spread across the row and have to be in the right order, only the name of the variable should
      be specified once in ``names``
    - ``names`` which are not specified as output variables are ignored
    - additional options are passed directly to ``numpy.genfromtxt()`
    """
    dtype = [(name, float, data.dtype[name].shape if name in data.dtype.names else ())
             for name in self.config['names']]
    try:
        raw = np.genfromtxt(self.config['path'], dtype=dtype, **self.config['options'])
    except OSError:
        self.logger.error(f'output file {self.config["path"]} not found')
        self.logger.info(f'cwd = {os.getcwd()}')
        dirname = os.path.dirname(self.config['path']) or '.'
        self.logger.info(f'ls {dirname} = {os.listdir(dirname)}')
        raise
    for key in self.config['names']:
        if key in data.dtype.names:
            data[key] = raw[key]


# === HDF5 Postprocessor === #


@Postprocessor.wrap('hdf5')
def HDF5Postprocessor(self, data):
    """ Postprocessor to read output from a HDF5 file

        - variables are assumed to be stored with the correct key and able to be converted immediately
        - not extensively tested
    """
    import h5py
    with h5py.File(self.config['path'], 'r') as f:
        for key in f.keys():
            data[key] = f[key]



"""
For computationally expensive simulations or experiments it is crucial to get the most information out of every
training point. This is not the case in the standard procedure of randomly selecting the training points.
In order to get the most out of the least number of training points, the next point is inferred by calculating an
acquisition function like the minimization of local variance or expected improvement.
"""
import numpy as np
from abc import abstractmethod
from warnings import warn

from profit.util.base_class import CustomABC
from profit.defaults import active_learning as defaults


class ActiveLearning(CustomABC):
    """Active learning base class.

    Parameters:
        runner (profit.run.Runner): Runner to dynamically start runs.
        variables (profit.util.variable.VariableGroup): Variables.
        ntrain (int): Total number of training points.
        nwarmup (int): Number of warmup (random) initialization points.
        batch_size (int): Number of training samples learned in parallel.
        convergence_criterion (float): AL is stopped when the loss of the acquisition function is lower than this
            criterion. Not implemented yet.
        nsearch (int): Number of possible candidate points in each dimension.
        make_plot (bool): Flat indicating if the AL progress is plotted.

    Attributes:
        krun (int): Current training cycle.
    """
    labels = {}

    def __init__(self, runner, variables, ntrain, nwarmup=defaults['nwarmup'], batch_size=defaults['batch_size'],
                 convergence_criterion=defaults['convergence_criterion'], nsearch=defaults['nsearch'],
                 make_plot=defaults['make_plot']):
        self.runner = runner
        self.variables = variables

        self.ntrain = ntrain
        self.nwarmup = min(nwarmup, ntrain)

        if nwarmup > ntrain:
            message = "nwarmup > ntrain. Setting nwarmup=ntrain."
            warn(message)

        self.batch_size = batch_size
        if (ntrain - nwarmup) % batch_size:
            raise RuntimeError("Number of learning points ({}) and batch size ({}) for AL not compatible!"
                               .format(ntrain - nwarmup, batch_size))

        self.convergence_criterion = convergence_criterion
        self.make_plot = make_plot
        self.krun = 0

    @abstractmethod
    def warmup(self, save_intermediate=defaults['save_intermediate']):
        """Warmup cycle before the actual learning starts."""
        pass

    @abstractmethod
    def learn(self, resume_from=defaults['resume_from'], save_intermediate=defaults['save_intermediate']):
        """Main loop for active learning."""
        pass

    def update_run(self, candidates):
        """Run a batch of simulations with the new candidates.

        Parameters:
            candidates (np.array): Input points to run the simulation on.
        """

        params_array = [{} for _ in range(self.batch_size)]

        for key, values in zip(self.variables.named_input.dtype.names, candidates.T):
            for idx, value in enumerate(values):
                params_array[idx][key] = value
        # Start batch
        self.runner.spawn_array(params_array, blocking=True)

    def update_data(self):
        """Update the variables with the runner data."""
        from profit.util import check_ndim

        for key in self.runner.input_data.dtype.names:
            self.variables[key].value = check_ndim(self.runner.interface.input[key])
        for key in self.runner.output_data.dtype.names:
            self.variables[key].value = check_ndim(self.runner.interface.output[key])

    @abstractmethod
    def save(self, path):
        """Save the AL model.

        Parameters:
            path (str): Path where the model is saved.
        """
        pass

    def save_intermediate(self, model_path=None, input_path=None, output_path=None):
        from profit.util.file_handler import FileHandler
        if model_path:
            self.save(model_path)
        if input_path:
            FileHandler.save(input_path, self.variables.named_input)
        if output_path:
            formatted_output_data = self.variables.formatted_output \
                if output_path.endswith('.txt') else self.variables.named_output
            FileHandler.save(output_path, formatted_output_data)
        print('Saved intermediate results.')

    @abstractmethod
    def plot(self):
        """Plot the progress of the AL learning."""
        pass

    @classmethod
    def from_config(cls, runner, variables, config, base_config):
        """Instantiates an ActiveLearning object from the configuration parameters.

        Parameters:
            runner (profit.run.runner.Runner): Runner instance.
            variables (profit.util.variable.VariableGroup): Variables.
            config (dict): Only the 'active_learning' part of the base_config.
            base_config (dict): The whole configuration parameters.

        Returns:
            profit.al.active_learning.ActiveLearning: AL instance.
        """

        child = cls[config['algorithm']['class']]
        child_instance = child.from_config(runner, variables, config, base_config)
        return child_instance



import numpy as np

from profit.al import ActiveLearning
from profit.defaults import active_learning as base_defaults
from profit.defaults import al_algorithm_mcmc as defaults

np.random.seed(43)

two = False


@ActiveLearning.register("mcmc")
class McmcAL(ActiveLearning):
    """Markov-chain Monte-Carlo active learning algorithm.

    Parameters:
        reference_data (np.ndarray): Observed experimental data points. This is not the simulated model data!
        warmup_cycles (int): Number of warmup cycles with `nwarmup` iterations each.
        target_acceptance_rate (float): Target rate with which probability new points are accepted.
        sigma_n (float): Estimated standard deviation of the experimental data.
        initial_points (list of float): Starting points for the MCMC.
        delayed_acceptancd (bool): Whether to use delayed acceptance with a surrogate model for the likelihood.

    Attributes:
        Xpred (np.ndarray): Matrix of the candidate points built with np.meshgrid.
        dx (np.ndarray): Distance between iterations in parameter space.
        ndim (int): Dimension of input parameters.
        Xtrain (np.ndarray): Array of sampled MCMC points.
        log_likelihood (np.ndarray): Array of the log likelihood during training.
        accepted (np.ndarray[bool]): Boolean array of accepted/rejected sample MCMC points.
    """
    labels = {}

    def __init__(self, runner, variables, reference_data, ntrain, warmup_cycles=defaults['warmup_cycles'],
                 nwarmup=base_defaults['nwarmup'], batch_size=base_defaults['batch_size'],
                 target_acceptance_rate=defaults['target_acceptance_rate'],
                 convergence_criterion=base_defaults['convergence_criterion'], nsearch=base_defaults['nsearch'],
                 sigma_n=defaults['sigma_n'], make_plot=base_defaults['make_plot'],
                 initial_points=defaults['initial_points'], save=defaults['save'], last_percent=defaults['last_percent'],
                 delayed_acceptance=defaults['delayed_acceptance']):

        super().__init__(runner, variables, ntrain, nwarmup, batch_size, convergence_criterion, nsearch, make_plot)
        self.reference_data = reference_data.reshape(1, -1)
        self.warmup_cycles = warmup_cycles
        self.target_acceptance_rate = target_acceptance_rate
        self.sigma_n = sigma_n
        self.initial_points = initial_points
        self.save_path = save
        self.last_index = int(last_percent * ntrain)

        al_keys = [var.name for var in variables.list if var.kind.lower() == 'activelearning']
        Xpred = [np.linspace(*var.constraints, nsearch) if var.name in al_keys else np.unique(var.value)
                 for var in variables.input_list]
        self.Xpred = np.hstack([xi.flatten().reshape(-1, 1) for xi in np.meshgrid(*Xpred)])

        self.ndim = self.Xpred.shape[-1]
        self.dx = np.full((self.ntrain, self.ndim), np.nan)
        self.Xtrain = np.full((self.ntrain, self.ndim), np.nan)
        self.ytrain = np.full((self.ntrain, self.reference_data.shape[-1]), np.nan)

        self.log_likelihood = np.full((self.ntrain, 1), np.nan)
        self.accepted = np.zeros((self.ntrain, self.ndim), dtype=bool)
        self.log_random = np.log(np.random.random((self.ntrain, self.ndim)))

        interface_class = self.runner.interface.__class__
        RunnerInterface = interface_class(self.runner.run_config['interface'], ntrain * self.ndim + 1,
                                          self.runner.base_config['input'], self.runner.base_config['output'],
                                          logger_parent=self.runner.logger)
        self.runner.interface = RunnerInterface

        if delayed_acceptance:
            from profit.sur import Surrogate
            self.delayed_acceptance_surrogate = Surrogate['GPy']()
            self.accepted_sur = np.zeros((self.ntrain, self.ndim), dtype=bool)
            self.log_random_sur = np.log(np.random.random((self.ntrain, self.ndim)))
        else:
            self.delayed_acceptance_surrogate = None

    def cost(self, y):
        return np.sum((y - self.reference_data)**2) / (y.shape[-1] * 2 * self.sigma_n**2)

    def f(self, x):
        nt = 250
        t = np.linspace(0, 2 * (1 - 1 / nt), nt)
        y_model = x[:,0] * np.sin((t - x[:, 1])**3)
        return y_model

    def warmup(self, save_intermediate=base_defaults['save_intermediate']):
        """Warmup MCMC."""
        from time import time
        self.Xtrain[0] = self.Xpred[np.random.choice(self.Xpred.shape[0])] \
            if self.initial_points is None else self.initial_points

        # TODO: Implement warmup with default AL model
        """ 
        if self.delayed_acceptance_surrogate:
            from profit.al.default_al import DefaultAL
            from profit.util.variable import OutputVariable
            log_likelihood_variable = OutputVariable('ll', 'Output', (self.ntrain, 1))
            self.variables.add(log_likelihood_variable)
            warmup_al = DefaultAL(self.runner, self.variables, self.delayed_acceptance_surrogate,
                                  self.nwarmup, 3, acquisition_function='log_likelihood')
            warmup_al.acquisition_function.reference_data = self.reference_data
            warmup_al.warmup()
            warmup_al.learn()
        """
        if two:
            self.ytrain[0] = self.f(self.Xtrain[[0]])
        else:
            self.update_run(self.Xtrain[[0]])
            self.ytrain[0] = self.runner.flat_output_data[0]

        self.log_likelihood[0] = -self.cost(self.ytrain[[0]])

        self.dx = np.random.randn(self.ntrain, self.ndim) * self.sigma_n * self.Xtrain[0]

        st = time()
        for cycle in range(self.warmup_cycles):
            self.krun = 1
            self.runner.next_run_id = 1
            self.accepted[:] = 0
            self.log_random[:] = np.log(np.random.random((self.ntrain, self.ndim)))

            self.do_mcmc(range(1, self.nwarmup+1))

            if self.make_plot:
                self.plot_mcmc('warmup')

            acceptance_rate = np.mean(self.accepted[:self.nwarmup], axis=0)
            print(f"Acceptance rate for warmup cycle {cycle+1}: {acceptance_rate}")
            self.dx = self.dx * np.exp(acceptance_rate / self.target_acceptance_rate - 1)

            if self.delayed_acceptance_surrogate:
                self.delayed_acceptance_surrogate.train(self.Xtrain[:self.nwarmup + 1],
                                                        self.log_likelihood[:self.nwarmup + 1])

            if save_intermediate:
                self.save_intermediate(**save_intermediate)

            if cycle+1 < self.warmup_cycles:
                self.Xtrain[0] = self.Xtrain[self.nwarmup]
                self.ytrain[0] = self.ytrain[self.nwarmup]
                self.log_likelihood[0] = self.log_likelihood[self.nwarmup]
        print("Runtime warmup: {}".format(time() - st))

    def learn(self, resume_from=base_defaults['resume_from'], save_intermediate=base_defaults['save_intermediate']):
        from time import time

        st = time()

        kstart = resume_from if resume_from is not None else self.nwarmup
        self.do_mcmc(range(kstart, self.ntrain))
        if self.delayed_acceptance_surrogate:
            print("Surrogate acceptance: ", np.mean(self.accepted_sur[kstart+1:], axis=0))

        if self.make_plot:
            self.plot_mcmc('learn')

        last_Xtrain = self.Xtrain[-self.last_index:]
        mean_Xtrain = np.mean(last_Xtrain, axis=0)
        std_Xtrain = np.std(last_Xtrain, axis=0)
        print("Best parameters: {} +- {}".format(mean_Xtrain, std_Xtrain))

        self.update_data()

        if save_intermediate:
            self.save_intermediate(**save_intermediate)

        print("Runtime main loop: {}".format(time() - st))
        if self.make_plot:
            from matplotlib.pyplot import show
            show()

    def do_mcmc(self, rng):
        from tqdm import tqdm
        for i in tqdm(rng):
            self.Xtrain[i] = self.Xtrain[i - 1]
            self.ytrain[i] = self.ytrain[i- 1]
            self.log_likelihood[i] = self.log_likelihood[i - 1]
            for col in range(self.ndim):
                Xguess = self.Xtrain[[i]].copy()
                Xguess[:, col] += self.dx[i- 1, col]

                if self.delayed_acceptance_surrogate and self.delayed_acceptance_surrogate.trained:
                    log_likelihood_guess_sur, _ = self.delayed_acceptance_surrogate.predict(Xguess)
                    A = log_likelihood_guess_sur - self.log_likelihood[i]
                    if A >= self.log_random_sur[i, col]:
                        self.accepted_sur[i - 1, col] = True
                    else:
                        continue

                if two:
                    y_guess = self.f(Xguess)
                else:
                    self.update_run(Xguess)
                    y_guess = self.runner.flat_output_data[[self.runner.next_run_id - 1]].copy()
                log_likelihood_guess = -self.cost(y_guess)
                A = log_likelihood_guess - self.log_likelihood[i]
                if A >= self.log_random[i, col]:
                    self.Xtrain[i] = Xguess
                    self.ytrain[i] = y_guess
                    self.log_likelihood[i] = log_likelihood_guess
                    self.accepted[i - 1, col] = True
            self.krun += 1

    def update_run(self, candidates):
        super().update_run(candidates)

    def update_data(self):
        """Update the variables with the runner data."""

        for idx, key in enumerate(self.runner.input_data.dtype.names):
            self.variables[key].value = self.Xtrain[:, [idx]]
        for idx, key in enumerate(self.runner.output_data.dtype.names):
            # TODO: Multi Output
            self.variables[key].value = self.ytrain

    def save(self, path):
        from profit.util.file_handler import FileHandler
        from os.path import dirname, join
        dirpath = dirname(path)
        FileHandler.save(join(dirpath, 'log_likelihood.txt'), self.log_likelihood)
        self.save_stats(join(dirpath, 'mcmc_stats.txt'))

    def save_stats(self, path):
        """Save mean and std of X values"""
        from profit.util.file_handler import FileHandler
        last_Xtrain = self.Xtrain[-self.last_index:]
        mean_Xtrain = np.mean(last_Xtrain, axis=0)
        std_Xtrain = np.std(last_Xtrain, axis=0)
        params = np.empty((2, 1), dtype=[('Xmean', 'float'), ('Xstd', 'float')])
        params['Xmean'] = mean_Xtrain.reshape(-1, 1)
        params['Xstd'] = std_Xtrain.reshape(-1, 1)
        FileHandler.save(path, params)

    def plot(self):
        import matplotlib.pyplot as plt
        plt.plot(self.reference_data.T, color='r')
        plt.plot(self.ytrain[-self.last_index:].T)
        plt.show()

    def plot_mcmc(self, phase):
        from matplotlib.pyplot import subplots

        fig, ax = subplots(3, 2)
        if phase == 'warmup':
            ind = range(0, self.nwarmup)
        elif phase == 'learn':
            ind = range(self.nwarmup, self.ntrain)
        else:
            ind = -1
        ax[0, 0].plot(self.ytrain[ind].T, 'b')
        ax[0, 0].plot(self.reference_data.T, 'r')
        ax[1, 0].hist(self.dx[ind])

        if self.ndim == 1:
            ax[0, 1].scatter(self.Xtrain[ind], self.log_likelihood[ind])
        elif self.ndim == 2:
            ax[0, 1].scatter(self.Xtrain[ind, 0], self.Xtrain[ind, 1],
                             c=self.log_likelihood[ind])

        ax[1, 1].plot(self.log_likelihood[ind])
        ax[2, 0].hist(self.Xtrain[ind])
        ax[2, 1].plot(self.Xtrain[ind])

    @classmethod
    def from_config(cls, runner, variables, config, base_config):
        from profit.util.file_handler import FileHandler
        reference_data = FileHandler.load(config['algorithm']['reference_data'])
        reference_data = np.hstack([reference_data[key] for key in reference_data.dtype.names])
        return cls(runner, variables, reference_data, base_config['ntrain'],
                   warmup_cycles=config['algorithm']['warmup_cycles'],
                   nwarmup=config['nwarmup'], batch_size=config['batch_size'],
                   target_acceptance_rate=config['algorithm']['target_acceptance_rate'],
                   convergence_criterion=config['convergence_criterion'], nsearch=config['nsearch'],
                   sigma_n=config['algorithm']['sigma_n'], make_plot=config['make_plot'],
                   initial_points=config['algorithm']['initial_points'],
                   save=config['algorithm']['save'], last_percent=config['algorithm']['last_percent'],
                   delayed_acceptance=config['algorithm']['delayed_acceptance'])



"""proFit default runner

in development (Mar 2021)
Goal: a class to manage and deploy runs
"""

import os
import sys
import logging
from abc import abstractmethod
from profit.util.base_class import CustomABC  # Abstract Base Class

from profit.util import load_includes, params2map

import numpy as np


class RunnerInterface(CustomABC):
    labels = {}
    internal_vars = [('DONE', np.bool8), ('TIME', np.uint32)]

    def __init__(self, config, size, input_config, output_config, *, logger_parent: logging.Logger = None):
        self.config = config  # base_config['run']['interface']
        self.logger = logging.getLogger('Runner Interface')
        if logger_parent is not None:
            self.logger.parent = logger_parent

        self.input_vars = [(variable, spec['dtype'].__name__) for variable, spec in input_config.items()]
        self.output_vars = [(variable, spec['dtype'].__name__, () if spec['size'] == (1, 1) else (spec['size'][-1],))
                            for variable, spec in output_config.items()]

        self.input = np.zeros(size, dtype=self.input_vars)
        self.output = np.zeros(size, dtype=self.output_vars)
        self.internal = np.zeros(size, dtype=self.internal_vars)
    
    def resize(self, size):
        if size <= self.size:
            self.logger.warning('shrinking RunnerInterface is not supported')
            return
        self.input.resize(size, refcheck=True)  # filled with 0 by default
        self.output.resize(size, refcheck=True)
        self.internal.resize(size, refcheck=True)
        
    @property
    def size(self):
        assert self.input.size == self.output.size == self.internal.size
        return self.input.size

    def poll(self):
        self.logger.debug('polling')

    def clean(self):
        self.logger.debug('cleaning')


# === Runner === #


class Runner(CustomABC):
    labels = {}

    # for now, implement the runner straightforward with less overhead
    # restructuring is always possible
    def __init__(self, interface_class, run_config, base_config, handle_logging=True):
        self.base_config = base_config
        self.run_config = run_config
        self.config = self.run_config['runner']
        self.logger = logging.getLogger('Runner')
        if handle_logging:
            log_handler = logging.FileHandler('runner.log', mode='w')
            log_formatter = logging.Formatter('{asctime} {levelname:8s} {name}: {message}', style='{')
            log_handler.setFormatter(log_formatter)
            self.logger.addHandler(log_handler)

            log_handler2 = logging.StreamHandler(sys.stderr)
            log_handler2.setFormatter(log_formatter)
            log_handler2.setLevel(logging.WARNING)
            self.logger.addHandler(log_handler2)
            self.logger.propagate = False
        if self.run_config['debug']:
            self.logger.setLevel(logging.DEBUG)
        self.interface: RunnerInterface = interface_class(self.run_config['interface'], self.base_config['ntrain'],
                                                          self.base_config['input'], self.base_config['output'],
                                                          logger_parent=self.logger)

        self.runs = {}  # run_id: (whatever data the system tracks)
        self.failed = {}  # ~runs, saving those that failed
        self.next_run_id = 0
        self.env = os.environ.copy()
        self.env['PROFIT_BASE_DIR'] = self.base_config['base_dir']
        self.env['PROFIT_CONFIG_PATH'] = base_config['config_path']  # ToDo better way to pass this?

    @classmethod
    def from_config(cls, config, base_config):
        child = cls[config['runner']['class']]
        interface_class = RunnerInterface[config['interface']['class']]
        return child(interface_class, config, base_config)

    def fill(self, params_array, offset=0):
        if offset + len(params_array) - 1 >= self.interface.size:
            self.interface.resize(max(offset + len(params_array), 2 * self.interface.size))
        for r, row in enumerate(params_array):
            mapping = params2map(row)
            for key, value in mapping.items():
                self.interface.input[key][r + offset] = value

    def fill_output(self, named_output, offset=0):
        if offset + len(named_output) - 1 >= self.interface.size:
            self.interface.resize(max(offset + len(named_output), 2 * self.interface.size))
        for r, row in enumerate(named_output):
            for key in row.dtype.names:
                self.interface.output[key][r + offset] = row[key]

    @abstractmethod
    def spawn_run(self, params=None, wait=False):
        """spawn a single run

        :param params: a mapping which defines input parameters to be set
        :param wait: whether to wait for the run to complete
        """
        if self.next_run_id >= self.interface.size:
            self.interface.resize(2 * self.interface.size)
        mapping = params2map(params)
        for key, value in mapping.items():
            self.interface.input[key][self.next_run_id] = value

    def spawn_array(self, params_array, blocking=True):
        if not blocking:
            raise NotImplementedError
        for params in params_array:
            self.spawn_run(params, wait=True)

    @abstractmethod
    def check_runs(self):
        pass

    def check_backup(self, run_id):
        """ preliminary checking for backup output

        ToDo: rework
        Works only with ZeroMQ Interface for now
        """
        import numpy as np
        path = os.path.join(self.base_config['run_dir'], f'run_{run_id:03d}')
        try:
            with open(os.path.join(path, 'profit_time.txt'), 'r') as file:
                self.interface.internal['TIME'] = int(file.read())
        except Exception as e:
            self.logger.debug(f'check time for run {run_id}: {e}')
        try:
            self.interface.output[run_id] = np.load(os.path.join(path, 'profit_results.npy'))
            self.interface.internal['DONE'][run_id] = True
        except Exception as e:
            self.logger.debug(f'check data for run {run_id}: {e}')
        return self.interface.internal['DONE'][run_id]

    @abstractmethod
    def cancel_all(self):
        pass

    def clean(self):
        self.interface.clean()

    @property
    def input_data(self):
        return self.interface.input[self.interface.internal['DONE']]

    @property
    def output_data(self):
        return self.interface.output[self.interface.internal['DONE']]

    @property
    def flat_output_data(self):
        from profit.util import flatten_struct
        return flatten_struct(self.output_data)


r"""This module contains the backends for various Gaussian Process surrogate models.

Gaussian Processes (GPs) are a generalization of Gaussian Distributions which are described by mean- and covariance
functions.
They can be used as as a non-parametric, supervised machine-learning technique for regression and classification.
The advantages of GPs over other machine-learning techniques such as Artificial Neural Networks is the consistent,
analytic mathematical derivation within probability-theory and therefore their intrinsic uncertainty-quantification of
the predictions by means of the covariance function. This makes the results of GP fits intuitive to interpret.

GPs belong (like e.g. Support Vector Machines) to the kernel methods of machine learning. The mean function is often
neglected and set to zero, because most functions can be modelled with a suitable covariance function or a combination
of several covariance functions. The most important kernels are the Gaussian (RBF) and its generalization,
the Matern kernels.

Isotropic RBF Kernel:
$$
\begin{align}
k(x, x') &= \sigma_f^2 \exp(-\frac{1}{2} \frac{\lvert x-x' \rvert}{l^2})
\end{align}
$$

Literature:
    Rasmussen & Williams 2006: General Introduction to Gaussian Processes
    Garnett 2014: Active Learning and Linear Embeddings
    Osborne 2012: Active Learning of Hyperparameters
"""

from abc import abstractmethod
import numpy as np
from profit.sur.sur import Surrogate
from profit.defaults import fit_gaussian_process as defaults, fit as base_defaults


class GaussianProcess(Surrogate):
    r"""This is the base class for all Gaussian Process models.

    Attributes:
        trained (bool): Flag that indicates if the model is already trained and ready to make predictions.
        fixed_sigma_n (bool/float/ndarray): Indicates if the data noise should be optimized or not.
            If an ndarray is given, its length must match the training data.
        Xtrain (ndarray): Input training points.
        ytrain (ndarray): Observed output data.
            Vector output is supported for independent variables only.
        ndim (int): Dimension of input data.
        output_ndim (int): Dimension of output data.
        kernel (str/object): Kernel identifier such as 'RBF' or directly the (surrogate specific) kernel object.
            Defaults to 'RBF'.
        hyperparameters (dict): Parameters like length-scale, variance and noise which can be optimized during training.
            As default, they are inferred from the training data.

    Default parameters:
        surrogate: GPy
        kernel: RBF

    Default hyperparameters:
        $l$ ... length scale
        $\sigma_f$ ... scaling
        $\sigma_n$ ... data noise

        $$
        \begin{align}
        l &= \frac{1}{2} \overline{\lvert x - x' \rvert} \\
        \sigma_f &= \overline{std(y)} \\
        \sigma_n &= 0.01 \cdot \overline{max(y) - min(y)}
        \end{align}
        $$
    """

    def __init__(self):
        super().__init__()
        self.kernel = None
        self.hyperparameters = {}

    def pre_train(self, X, y, kernel=defaults['kernel'], hyperparameters=defaults['hyperparameters'],
                  fixed_sigma_n=base_defaults['fixed_sigma_n']):
        """Check the training data, initialize the hyperparameters and set the kernel either from the given parameter,
        from config or from the default values.

        Parameters:
            X: (n, d) array of input training data.
            y: (n, D) array of training output.
            kernel (str/object): Identifier of kernel like 'RBF' or directly the kernel object of the
                specific surrogate.
            hyperparameters (dict): Hyperparameters such as length scale, variance and noise.
                Taken either from given parameter, config file or inferred from the training data.
                The hyperparameters can be different depending on the kernel. E.g. The length scale can be a scalar,
                a vector of the size of the training data, or for the custom LinearEmbedding kernel a matrix.
            fixed_sigma_n (bool/float/ndarray): Indicates if the data noise should be optimized or not.
                If an ndarray is given, its length must match the training data.
        """

        self.Xtrain, self.ytrain = np.atleast_2d(X), np.atleast_2d(y)

        # Set attributes either from config, the given parameters or from defaults
        self.set_attributes(fixed_sigma_n=fixed_sigma_n, kernel=kernel, hyperparameters=hyperparameters)

        self.encode_training_data()
        self.ndim = self.Xtrain.shape[-1]

        # Set hyperparameter inferred values
        inferred_hyperparameters = self.infer_hyperparameters()
        for key, value in self.hyperparameters.items():
            if value is None:
                value = inferred_hyperparameters[key]
            self.hyperparameters[key] = np.atleast_1d(value)
        self.print_hyperparameters("Initial")

        # Convert a kernel identifier as string to the class object
        if isinstance(self.kernel, str):
            self.kernel = self.select_kernel(self.kernel)

    def set_attributes(self, **kwargs):
        for key, value in kwargs.items():
            if not getattr(self, key):
                new_value = defaults.get(key, value) if not value else value
                if isinstance(new_value, dict):
                    new_value = new_value.copy()
                setattr(self, key, new_value)

    def infer_hyperparameters(self):
        # Infer hyperparameters from data
        dist = np.linalg.norm(self.Xtrain[:, None, :] - self.Xtrain[None, :, :], axis=-1)
        std = np.std(self.ytrain, axis=0)
        spread = np.abs(self.ytrain.max(axis=0) - self.ytrain.min(axis=0))
        if not hasattr(self, 'ARD'):
            dist, std, spread = np.mean(dist), np.mean(std), np.mean(spread)
        length_scale = np.array([0.5 * dist])
        sigma_f = np.array([std])
        sigma_n = 1e-2 * np.array([spread])
        return {'length_scale': length_scale, 'sigma_f': sigma_f, 'sigma_n': sigma_n}

    @abstractmethod
    def train(self, X, y, kernel=defaults['kernel'], hyperparameters=defaults['hyperparameters'],
              fixed_sigma_n=base_defaults['fixed_sigma_n'], return_hess_inv=False):
        """Trains the model on the dataset.

        After initializing the model with a kernel function and initial hyperparameters,
        it can be trained on input data X and observed output data y by optimizing the model's hyperparameters.
        This is done by minimizing the negative log likelihood.

        Parameters:
            X (ndarray): (n, d) array of input training data.
            y (ndarray): (n, D) array of training output.
            kernel (str/object): Identifier of kernel like 'RBF' or directly the kernel object of the surrogate.
            hyperparameters (dict): Hyperparameters such as length scale, variance and noise.
                Taken either from given parameter, config file or inferred from the training data.
                The hyperparameters can be different depending on the kernel. E.g. The length scale can be a scalar,
                a vector of the size of the training data, or for the custom LinearEmbedding kernel a matrix.
            fixed_sigma_n (bool): Indicates if the data noise should be optimized or not.
            return_hess_inv (bool): Whether to the attribute hess_inv after optimization. This is important
                for active learning.
        """
        pass

    def post_train(self):
        self.trained = True
        self.decode_training_data()

    def pre_predict(self, Xpred):
        """Prepares the surrogate for prediction by checking if it is trained and validating the data.

        Parameters:
            Xpred (ndarray): Input points for prediction

        Returns:
            ndarray: Checked input data or default values inferred from training data.
        """
        from profit.util import check_ndim

        if not self.trained:
            raise RuntimeError("Need to train() before predict()!")

        if Xpred is None:
            Xpred = self.default_Xpred()
        Xpred = check_ndim(Xpred)
        Xpred = self.encode_predict_data(Xpred)
        return Xpred

    @abstractmethod
    def predict(self, Xpred, add_data_variance=True):
        r"""Predicts the output at test points Xpred.

        Parameters:
            Xpred (ndarray/list): Input points for prediction.
            add_data_variance (bool): Adds the data noise $\sigma_n^2$ to the prediction variance.
                This is especially useful for plotting.

        Returns:
            tuple: a tuple containing:
                - ymean (ndarray) Predicted output values at the test input points.
                - yvar (ndarray): Diagonal of the predicted covariance matrix.
        """
        pass

    @abstractmethod
    def optimize(self, **opt_kwargs):
        """Find optimized hyperparameters of the model. Optional kwargs for tweaking optimization.

        Parameters:
            opt_kwargs: Keyword arguments for optimization.
        """
        pass

    @classmethod
    def from_config(cls, config, base_config):
        """Instantiate a GP model from the configuration file with kernel and hyperparameters.

        Parameters:
            config (dict): Only the 'fit' part of the base_config.
            base_config (dict): The whole configuration parameters.

        Returns:
            profit.sur.gaussian_process.GaussianProcess: Instantiated surrogate.
        """

        self = cls()
        self.kernel = config['kernel']
        self.hyperparameters = config['hyperparameters']
        return self

    def select_kernel(self, kernel):
        """Convert the name of the kernel as string to the kernel class object of the surrogate.

        Parameters:
            kernel (str): Kernel string such as 'RBF' or depending on the surrogate also product and sum kernels
                such as 'RBF+Matern52'.
        Returns:
            object: Custom or imported kernel object. This is the function which builds the kernel and not
            the calculated covariance matrix.
        """
        pass

    def decode_hyperparameters(self):
        """Decodes the hyperparameters, as encoded ones are used in the surrogate model."""
        for key, value in self.hyperparameters.items():
            new_value = value
            if key == 'length_scale':
                for enc in self.input_encoders[::-1]:
                    new_value = enc.decode_hyperparameters(new_value)
            if key in ('sigma_f', 'sigma_n'):
                for enc in self.output_encoders[::-1]:
                    new_value = enc.decode_hyperparameters(new_value)
            new_value = self.special_hyperparameter_decoding(key, new_value)
            self.hyperparameters[key] = new_value

    def special_hyperparameter_decoding(self, key, value):
        return np.atleast_1d(value)

    def print_hyperparameters(self, prefix):
        """Helper function to print the hyperparameter dict.

        Parameters:
            prefix (str): Usually 'Initialized', 'Loaded' or 'Optimized' to identify the state of the hyperparameters.
        """

        #hyperparameter_str = ["{} hyperparameters:".format(prefix)]
        #hyperparameter_str += ["{k}: {v}".format(k=key, v=value) for key, value in self.hyperparameters.items()]
        #print('\n'.join(hyperparameter_str))
        pass  # TODO: Include in DEBUG logging.



"""
Testcases for encoders:
- Exclude
- Log10
- Normalization

Separate linear-reduction tests:
- PCA
- Karuhnen-Loeve
"""
import numpy as np
from profit.sur.encoders import Encoder


def test_exclude():

    COLUMNS= [2]
    CONFIG = {'class': 'Exclude', 'columns': COLUMNS, 'parameters': {}}
    SIZE = (10, 4)
    n = SIZE[0] * SIZE[1]
    X = np.linspace(0, n-1, n).reshape(SIZE)

    enc = Encoder['Exclude'](COLUMNS)
    assert enc.repr == CONFIG
    X_enc = enc.encode(X)
    assert np.all(X_enc == X[:, [0, 1, 3]])
    X_dec = enc.decode(X_enc)
    assert np.all(X_dec == X)


def test_log10():
    from profit.sur.encoders import Log10Encoder

    COLUMNS= [2, 3]
    CONFIG = {'class': 'Log10', 'columns': COLUMNS, 'parameters': {}}
    SIZE = (10, 4)
    n = SIZE[0] * SIZE[1]
    X = np.linspace(0, n-1, n).reshape(SIZE)
    X_log = X.copy()
    X_log[:, COLUMNS] = np.log10(X_log[:, COLUMNS])

    enc = Log10Encoder(COLUMNS)
    assert enc.repr == CONFIG
    X_enc = enc.encode(X)
    assert np.all(X_enc == X_log)
    X_dec = enc.decode(X_enc)
    assert np.allclose(X_dec, X, atol=1e-7)


def test_normalization():

    COLUMNS= [0, 1, 2, 3]
    CONFIG = {'class': 'Normalization', 'columns': COLUMNS, 'parameters': {}}
    SIZE = (10, 4)
    n = SIZE[0] * SIZE[1]
    X = np.linspace(0, n-1, n).reshape(SIZE)
    norm_box = np.hstack([np.linspace(0, 1, SIZE[0]).reshape(-1, 1) for _ in range(SIZE[1])])

    enc = Encoder['Normalization'](COLUMNS)
    assert enc.repr == CONFIG
    X_enc = enc.encode(X)
    assert np.allclose(X_enc, norm_box)
    X_dec = enc.decode(X_enc)
    assert np.allclose(X_dec, X)

    y = np.sin(X[:, [1]])
    enc = Encoder['Normalization']([0])
    y_enc = enc.encode(y)
    assert np.allclose(np.max(y_enc), 1)
    assert np.allclose(np.min(y_enc), 0)
    y_dec = enc.decode(y_enc)
    assert np.allclose(y_dec, y)

    hyperparameters = {'length_scale': np.array([1.3, 0.2, 2.4]),
                       'sigma_f': np.array([0.7]),
                       'sigma_n': np.array([0.05])}

    xmean = 1.7
    xstd = 2.8
    xmin = 0.5
    xmax = 3
    xmax_centered = (xmax - xmean) / xstd
    xmin_centered = (xmin - xmean) / xstd
    scaling = (xmax_centered - xmin_centered) * xstd

    enc = Encoder['Normalization'](COLUMNS, parameters={'xmean': xmean, 'xstd': xstd, 'xmin': xmin, 'xmax': xmax,
                                                        'xmin_centered': xmin_centered, 'xmax_centered': xmax_centered})
    length_scale_dec = enc.decode_hyperparameters(hyperparameters['length_scale'])
    sigma_f_dec = np.sqrt(enc.decode_variance(hyperparameters['sigma_f']**2))
    sigma_n_dec = np.sqrt(enc.decode_variance(hyperparameters['sigma_n']**2))
    assert np.allclose(length_scale_dec, hyperparameters['length_scale'] * scaling)
    assert np.allclose(sigma_f_dec, hyperparameters['sigma_f'] * scaling)
    assert np.allclose(sigma_n_dec, hyperparameters['sigma_n'] * scaling)
